@@ -28,20 +28,23 @@ conversions and pagination.
 
 """
 
-from typing import Any, Callable, Dict, Generator, List, Union
+from datetime import datetime, timedelta
+from typing import Any, Callable, Dict, Generator, List, Optional, Union
 from urllib.parse import urlparse
 
 import dateutil.parser
 import requests
 
-from swh.model.identifiers import \
-    SNAPSHOT, REVISION, RELEASE, DIRECTORY, CONTENT
+from swh.model.identifiers import SNAPSHOT, REVISION, RELEASE, DIRECTORY, CONTENT
 from swh.model.identifiers import PersistentId as PID
 from swh.model.identifiers import parse_persistent_identifier as parse_pid
 
+from .auth import AuthenticationError, OpenIDConnectSession, SWH_OIDC_SERVER_URL
 
 PIDish = Union[PID, str]
 
+ORIGIN_VISIT = "origin_visit"
+
 
 def _get_pid(pidish: PIDish) -> PID:
     """Parse string to PID if needed"""
@@ -60,6 +63,7 @@ def typify(data: Any, obj_type: str) -> Any:
     - timestamps are converted from strings to datetime.datetime objects
 
     """
+
     def to_pid(object_type, s):
         return PID(object_type=object_type, object_id=s)
 
@@ -67,43 +71,47 @@ def typify(data: Any, obj_type: str) -> Any:
         return dateutil.parser.parse(s)
 
     def obj_type_of_entry_type(s):
-        if s == 'file':
+        if s == "file":
             return CONTENT
-        elif s == 'dir':
+        elif s == "dir":
             return DIRECTORY
-        elif s == 'rev':
+        elif s == "rev":
             return REVISION
         else:
-            raise ValueError(f'invalid directory entry type: {s}')
+            raise ValueError(f"invalid directory entry type: {s}")
 
     if obj_type == SNAPSHOT:
         for name, target in data.items():
-            if target['target_type'] != 'alias':
+            if target["target_type"] != "alias":
                 # alias targets do not point to objects via PIDs; others do
-                target['target'] = to_pid(target['target_type'],
-                                          target['target'])
+                target["target"] = to_pid(target["target_type"], target["target"])
     elif obj_type == REVISION:
-        data['id'] = to_pid(obj_type, data['id'])
-        data['directory'] = to_pid(DIRECTORY, data['directory'])
-        for key in ('date', 'committer_date'):
+        data["id"] = to_pid(obj_type, data["id"])
+        data["directory"] = to_pid(DIRECTORY, data["directory"])
+        for key in ("date", "committer_date"):
             data[key] = to_date(data[key])
-        for parent in data['parents']:
-            parent['id'] = to_pid(REVISION, parent['id'])
+        for parent in data["parents"]:
+            parent["id"] = to_pid(REVISION, parent["id"])
     elif obj_type == RELEASE:
-        data['id'] = to_pid(obj_type, data['id'])
-        data['date'] = to_date(data['date'])
-        data['target'] = to_pid(data['target_type'], data['target'])
+        data["id"] = to_pid(obj_type, data["id"])
+        data["date"] = to_date(data["date"])
+        data["target"] = to_pid(data["target_type"], data["target"])
     elif obj_type == DIRECTORY:
         dir_pid = None
         for entry in data:
-            dir_pid = dir_pid or to_pid(obj_type, entry['dir_id'])
-            entry['dir_id'] = dir_pid
-            entry['target'] = to_pid(obj_type_of_entry_type(entry['type']),
-                                     entry['target'])
+            dir_pid = dir_pid or to_pid(obj_type, entry["dir_id"])
+            entry["dir_id"] = dir_pid
+            entry["target"] = to_pid(
+                obj_type_of_entry_type(entry["type"]), entry["target"]
+            )
     elif obj_type == CONTENT:
         pass  # nothing to do for contents
+    elif obj_type == ORIGIN_VISIT:
+        data['date'] = to_date(data['date'])
+        if data['snapshot'] is not None:
+            data['snapshot'] = to_pid(SNAPSHOT, data['snapshot'])
     else:
-        raise ValueError(f'invalid object type: {obj_type}')
+        raise ValueError(f"invalid object type: {obj_type}")
 
     return data
 
@@ -115,7 +123,11 @@ class WebAPIClient:
 
     """
 
-    def __init__(self, api_url='https://archive.softwareheritage.org/api/1'):
+    def __init__(
+        self,
+        api_url="https://archive.softwareheritage.org/api/1",
+        auth_url=SWH_OIDC_SERVER_URL,
+    ):
         """Create a client for the Software Heritage Web API
 
         See: https://archive.softwareheritage.org/api/
@@ -125,11 +137,13 @@ class WebAPIClient:
                 "https://archive.softwareheritage.org/api/1")
 
         """
-        api_url = api_url.rstrip('/')
+        api_url = api_url.rstrip("/")
         u = urlparse(api_url)
 
         self.api_url = api_url
         self.api_path = u.path
+        self.oidc_session = OpenIDConnectSession(oidc_server_url=auth_url)
+        self.oidc_profile: Optional[Dict[str, Any]] = None
 
         self._getters: Dict[str, Callable[[PIDish], Any]] = {
             CONTENT: self.content,
@@ -139,8 +153,9 @@ class WebAPIClient:
             SNAPSHOT: self._get_snapshot,
         }
 
-    def _call(self, query: str, http_method: str = 'get',
-              **req_args) -> requests.models.Response:
+    def _call(
+        self, query: str, http_method: str = "get", **req_args
+    ) -> requests.models.Response:
         """Dispatcher for archive API invocation
 
         Args:
@@ -156,16 +171,25 @@ class WebAPIClient:
         if urlparse(query).scheme:  # absolute URL
             url = query
         else:  # relative URL; prepend base API URL
-            url = '/'.join([self.api_url, query])
+            url = "/".join([self.api_url, query])
         r = None
 
-        if http_method == 'get':
-            r = requests.get(url, **req_args)
+        headers = {}
+        if self.oidc_profile is not None:
+            # use bearer token authentication
+            if datetime.now() > self.oidc_profile["expires_at"]:
+                # refresh access token if it has expired
+                self.authenticate(self.oidc_profile["refresh_token"])
+            access_token = self.oidc_profile["access_token"]
+            headers = {"Authorization": f"Bearer {access_token}"}
+
+        if http_method == "get":
+            r = requests.get(url, **req_args, headers=headers)
             r.raise_for_status()
-        elif http_method == 'head':
-            r = requests.head(url, **req_args)
+        elif http_method == "head":
+            r = requests.head(url, **req_args, headers=headers)
         else:
-            raise ValueError(f'unsupported HTTP method: {http_method}')
+            raise ValueError(f"unsupported HTTP method: {http_method}")
 
         return r
 
@@ -195,8 +219,7 @@ class WebAPIClient:
         pid_ = _get_pid(pid)
         return self._getters[pid_.object_type](pid_)
 
-    def iter(self, pid: PIDish, **req_args) -> Generator[Dict[str, Any],
-                                                         None, None]:
+    def iter(self, pid: PIDish, **req_args) -> Generator[Dict[str, Any], None, None]:
         """Stream over the information about an object of any kind
 
         Streaming variant of get()
@@ -215,7 +238,7 @@ class WebAPIClient:
         elif obj_type == CONTENT:
             yield from [self.content(pid_)]
         else:
-            raise ValueError(f'invalid object type: {obj_type}')
+            raise ValueError(f"invalid object type: {obj_type}")
 
     def content(self, pid: PIDish, **req_args) -> Dict[str, Any]:
         """Retrieve information about a content object
@@ -229,9 +252,11 @@ class WebAPIClient:
 
         """
         return typify(
-            self._call(f'content/sha1_git:{_get_pid(pid).object_id}/',
-                       **req_args).json(),
-            CONTENT)
+            self._call(
+                f"content/sha1_git:{_get_pid(pid).object_id}/", **req_args
+            ).json(),
+            CONTENT,
+        )
 
     def directory(self, pid: PIDish, **req_args) -> List[Dict[str, Any]]:
         """Retrieve information about a directory object
@@ -245,9 +270,9 @@ class WebAPIClient:
 
         """
         return typify(
-            self._call(f'directory/{_get_pid(pid).object_id}/',
-                       **req_args).json(),
-            DIRECTORY)
+            self._call(f"directory/{_get_pid(pid).object_id}/", **req_args).json(),
+            DIRECTORY,
+        )
 
     def revision(self, pid: PIDish, **req_args) -> Dict[str, Any]:
         """Retrieve information about a revision object
@@ -261,9 +286,9 @@ class WebAPIClient:
 
         """
         return typify(
-            self._call(f'revision/{_get_pid(pid).object_id}/',
-                       **req_args).json(),
-            REVISION)
+            self._call(f"revision/{_get_pid(pid).object_id}/", **req_args).json(),
+            REVISION,
+        )
 
     def release(self, pid: PIDish, **req_args) -> Dict[str, Any]:
         """Retrieve information about a release object
@@ -277,12 +302,13 @@ class WebAPIClient:
 
         """
         return typify(
-            self._call(f'release/{_get_pid(pid).object_id}/',
-                       **req_args).json(),
-            RELEASE)
+            self._call(f"release/{_get_pid(pid).object_id}/", **req_args).json(),
+            RELEASE,
+        )
 
-    def snapshot(self, pid: PIDish,
-                 **req_args) -> Generator[Dict[str, Any], None, None]:
+    def snapshot(
+        self, pid: PIDish, **req_args
+    ) -> Generator[Dict[str, Any], None, None]:
         """Retrieve information about a snapshot object
 
         Args:
@@ -300,12 +326,52 @@ class WebAPIClient:
         """
         done = False
         r = None
-        query = f'snapshot/{_get_pid(pid).object_id}/'
+        query = f"snapshot/{_get_pid(pid).object_id}/"
 
         while not done:
-            r = self._call(query, http_method='get', **req_args)
-            yield typify(r.json()['branches'], SNAPSHOT)
+            r = self._call(query, http_method="get", **req_args)
+            yield typify(r.json()["branches"], SNAPSHOT)
+            if "next" in r.links and "url" in r.links["next"]:
+                query = r.links["next"]["url"]
+            else:
+                done = True
+
+    def visits(self,
+               origin: str,
+               per_page: Optional[int] = None,
+               last_visit: Optional[int] = None,
+               **req_args) -> Generator[Dict[str, Any], None, None]:
+        """List visits of an origin
+
+        Args:
+            origin: the URL of a software origin
+            per_page: the number of visits to list
+            last_visit: visit to start listing from
+            req_args: extra keyword arguments for requests.get()
+
+        Returns:
+            an iterator over visits of the origin
+
+        Raises:
+            requests.HTTPError: if HTTP request fails
+
+        """
+        done = False
+        r = None
+
+        params = []
+        if last_visit is not None:
+            params.append(("last_visit", last_visit))
+        if per_page is not None:
+            params.append(("per_page", per_page))
+
+        query = f'origin/{origin}/visits/'
+
+        while not done:
+            r = self._call(query, http_method='get', params=params, **req_args)
+            yield from [typify(v, ORIGIN_VISIT) for v in r.json()]
             if 'next' in r.links and 'url' in r.links['next']:
+                params = []
                 query = r.links['next']['url']
             else:
                 done = True
@@ -321,8 +387,13 @@ class WebAPIClient:
           requests.HTTPError: if HTTP request fails
 
         """
-        return bool(self._call(f'content/sha1_git:{_get_pid(pid).object_id}/',
-                               http_method='head', **req_args))
+        return bool(
+            self._call(
+                f"content/sha1_git:{_get_pid(pid).object_id}/",
+                http_method="head",
+                **req_args,
+            )
+        )
 
     def directory_exists(self, pid: PIDish, **req_args) -> bool:
         """Check if a directory object exists in the archive
@@ -335,8 +406,11 @@ class WebAPIClient:
           requests.HTTPError: if HTTP request fails
 
         """
-        return bool(self._call(f'directory/{_get_pid(pid).object_id}/',
-                               http_method='head', **req_args))
+        return bool(
+            self._call(
+                f"directory/{_get_pid(pid).object_id}/", http_method="head", **req_args
+            )
+        )
 
     def revision_exists(self, pid: PIDish, **req_args) -> bool:
         """Check if a revision object exists in the archive
@@ -349,8 +423,11 @@ class WebAPIClient:
           requests.HTTPError: if HTTP request fails
 
         """
-        return bool(self._call(f'revision/{_get_pid(pid).object_id}/',
-                               http_method='head', **req_args))
+        return bool(
+            self._call(
+                f"revision/{_get_pid(pid).object_id}/", http_method="head", **req_args
+            )
+        )
 
     def release_exists(self, pid: PIDish, **req_args) -> bool:
         """Check if a release object exists in the archive
@@ -363,8 +440,11 @@ class WebAPIClient:
           requests.HTTPError: if HTTP request fails
 
         """
-        return bool(self._call(f'release/{_get_pid(pid).object_id}/',
-                               http_method='head', **req_args))
+        return bool(
+            self._call(
+                f"release/{_get_pid(pid).object_id}/", http_method="head", **req_args
+            )
+        )
 
     def snapshot_exists(self, pid: PIDish, **req_args) -> bool:
         """Check if a snapshot object exists in the archive
@@ -377,11 +457,13 @@ class WebAPIClient:
           requests.HTTPError: if HTTP request fails
 
         """
-        return bool(self._call(f'snapshot/{_get_pid(pid).object_id}/',
-                               http_method='head', **req_args))
+        return bool(
+            self._call(
+                f"snapshot/{_get_pid(pid).object_id}/", http_method="head", **req_args
+            )
+        )
 
-    def content_raw(self, pid: PIDish,
-                    **req_args) -> Generator[bytes, None, None]:
+    def content_raw(self, pid: PIDish, **req_args) -> Generator[bytes, None, None]:
         """Iterate over the raw content of a content object
 
         Args:
@@ -392,8 +474,35 @@ class WebAPIClient:
           requests.HTTPError: if HTTP request fails
 
         """
-        r = self._call(f'content/sha1_git:{_get_pid(pid).object_id}/raw/',
-                       stream=True, **req_args)
+        r = self._call(
+            f"content/sha1_git:{_get_pid(pid).object_id}/raw/", stream=True, **req_args
+        )
         r.raise_for_status()
 
         yield from r.iter_content(chunk_size=None, decode_unicode=False)
+
+    def authenticate(self, refresh_token: str):
+        """Authenticate API requests using OpenID Connect bearer token
+
+        Args:
+            refresh_token: A refresh token retrieved using the
+                ``swh auth login`` command (see :ref:`swh-web-client-auth`
+                section in main documentation)
+
+        Raises:
+            swh.web.client.auth.AuthenticationError: if authentication fails
+
+        """
+        now = datetime.now()
+        try:
+            self.oidc_profile = self.oidc_session.refresh(refresh_token)
+            assert self.oidc_profile
+            if "expires_in" in self.oidc_profile:
+                expires_in = self.oidc_profile["expires_in"]
+                expires_at = now + timedelta(seconds=expires_in)
+                self.oidc_profile["expires_at"] = expires_at
+        except Exception as e:
+            raise AuthenticationError(str(e))
+        if "access_token" not in self.oidc_profile:
+            # JSON error response
+            raise AuthenticationError(self.oidc_profile)
